#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import re
import uuid
from typing import List, Optional, Union

import dagster._check as check
from dagster import (
    Field,
    InitResourceContext,
    Noneable,
    StringSource,
    resource,
)
from dagster._core.storage.file_manager import check_file_like_obj
from dagster_gcp.gcs.file_manager import GCSFileHandle, GCSFileManager
from google.cloud import storage
from google.oauth2 import service_account

from ..config import get_public_url_for_gcs_file


class PublicGCSFileHandle(GCSFileHandle):
    @property
    def public_url(self):
        return get_public_url_for_gcs_file(self.gcs_bucket, self.gcs_key)


class ContentTypeAwareGCSFileManager(GCSFileManager):
    """
    Slighlty modified dagster_gcp.gcs.file_manager.GCSFileManager
    to allow setting the content type of the file
    """

    def get_content_type(self, ext):
        if ext == "csv":
            return "text/csv"
        elif ext == "json":
            return "application/json"
        elif ext == "html":
            return "text/html"
        elif ext == "md":
            return "text/markdown"
        else:
            return "text/plain"

    def get_full_key(self, *args, **kwargs):
        full_key = super().get_full_key(*args, **kwargs)

        # remove the first slash if it exists to prevent double slashes
        if full_key.startswith("/"):
            full_key = full_key[1:]

        return full_key

    def write(
        self, file_obj, mode="wb", ext=None, key: Optional[str] = None
    ) -> PublicGCSFileHandle:
        """
        Reworked from dagster_gcp.gcs.file_manager.GCSFileManager.write

        As the original method does not allow to set the content type of the file
        """
        key = check.opt_str_param(key, "key", default=str(uuid.uuid4()))
        check_file_like_obj(file_obj)
        gcs_key = self.get_full_key(key + (("." + ext) if ext is not None else ""))

        bucket_obj = self._client.bucket(self._gcs_bucket)
        blob = bucket_obj.blob(gcs_key)

        # Set Cache-Control header to no-cache to avoid caching issues
        # This is IMPORTANT because if we don't set this header, the metadata file will be cached by GCS
        # and the next time we try to download it, we will get the stale version
        blob.cache_control = "no-cache"
        blob.content_type = self.get_content_type(ext)

        blob.upload_from_file(file_obj)
        return PublicGCSFileHandle(self._gcs_bucket, gcs_key)

    def delete_by_key(
        self, key: str, ext: Optional[str] = None
    ) -> Optional[PublicGCSFileHandle]:
        gcs_key = self.get_full_key(key + (("." + ext) if ext is not None else ""))
        bucket_obj = self._client.bucket(self._gcs_bucket)
        blob = bucket_obj.blob(gcs_key)

        # if the file does not exist, return None
        if not blob.exists():
            return None

        blob.delete()
        return PublicGCSFileHandle(self._gcs_bucket, gcs_key)


@resource(config_schema={"gcp_gcs_cred_string": StringSource})
def gcp_gcs_client(resource_context: InitResourceContext) -> storage.Client:
    """Create a connection to gcs."""

    resource_context.log.info("retrieving gcp_gcs_client")
    gcp_gcs_cred_string = resource_context.resource_config["gcp_gcs_cred_string"]
    gcp_gsm_cred_json = json.loads(gcp_gcs_cred_string)
    credentials = service_account.Credentials.from_service_account_info(
        gcp_gsm_cred_json
    )
    return storage.Client(
        credentials=credentials,
        project=credentials.project_id,
    )


@resource(
    required_resource_keys={"gcp_gcs_client"},
    config_schema={"gcs_bucket": StringSource, "prefix": StringSource},
)
def gcs_file_manager(resource_context) -> GCSFileManager:
    """FileManager that provides abstract access to GCS.

    Implements the `dagster._core.storage.file_manager.FileManager` API.
    """

    storage_client = resource_context.resources.gcp_gcs_client

    return ContentTypeAwareGCSFileManager(
        client=storage_client,
        gcs_bucket=resource_context.resource_config["gcs_bucket"],
        gcs_base_key=resource_context.resource_config["prefix"],
    )


@resource(
    required_resource_keys={"gcp_gcs_client"},
    config_schema={
        "gcs_bucket": StringSource,
        "prefix": Noneable(StringSource),
        "gcs_filename": StringSource,
    },
)
def gcs_file_blob(resource_context: InitResourceContext) -> storage.Blob:
    """
    Create a connection to a gcs file blob.

    This is implemented so we are able to retrieve the metadata of a file
    before committing to downloading the file.
    """
    gcs_bucket = resource_context.resource_config["gcs_bucket"]
    storage_client = resource_context.resources.gcp_gcs_client

    bucket = storage_client.get_bucket(gcs_bucket)

    prefix = resource_context.resource_config["prefix"]
    gcs_filename = resource_context.resource_config["gcs_filename"]
    gcs_file_path = f"{prefix}/{gcs_filename}" if prefix else gcs_filename

    resource_context.log.info(
        f"retrieving gcs file blob {gcs_file_path} in bucket: {gcs_bucket}"
    )

    gcs_file_blob = bucket.get_blob(gcs_file_path)
    if not gcs_file_blob or not gcs_file_blob.exists():
        raise Exception(f"File does not exist at path: {gcs_file_path}")

    return gcs_file_blob


@resource(
    required_resource_keys={"gcp_gcs_client"},
    config_schema={
        "gcs_bucket": StringSource,
        "prefix": StringSource,
        "match_regex": StringSource,
        "only_one": Field(config=bool, default_value=False),
        "sort_key": Field(config=str, is_required=False),
        "reverse_sort": Field(config=bool, default_value=False),
    },
)
def gcs_directory_blobs(
    resource_context: InitResourceContext,
) -> Union[List[storage.Blob], storage.Blob]:
    """
    List all blobs in a bucket that match the prefix.
    """
    gcs_bucket = resource_context.resource_config["gcs_bucket"]
    prefix = resource_context.resource_config["prefix"]
    match_regex = resource_context.resource_config["match_regex"]
    only_one = resource_context.resource_config["only_one"]
    sort_key = resource_context.resource_config.get("sort_key")
    reverse_sort = resource_context.resource_config["reverse_sort"]

    storage_client = resource_context.resources.gcp_gcs_client
    bucket = storage_client.get_bucket(gcs_bucket)

    resource_context.log.info(
        f"retrieving gcs file blobs for prefix: {prefix}, match_regex: {match_regex}, in bucket: {gcs_bucket}"
    )

    gcs_file_blobs = bucket.list_blobs(prefix=prefix)
    if match_regex:
        gcs_file_blobs = [
            blob for blob in gcs_file_blobs if re.match(match_regex, blob.name)
        ]

    if sort_key:
        gcs_file_blobs = sorted(
            gcs_file_blobs, key=lambda x: getattr(x, sort_key), reverse=reverse_sort
        )

    if only_one:
        return gcs_file_blobs[0] if gcs_file_blobs else None

    return gcs_file_blobs



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import base64
import hashlib

import yaml
from dagster import OpExecutionContext, Output, asset
from github import Repository

from ..logging import sentry
from ..models.metadata import LatestMetadataEntry, MetadataDefinition

GROUP_NAME = "github"


def _get_md5_of_github_file(
    context: OpExecutionContext, github_connector_repo: Repository, path: str
) -> str:
    """
    Return the md5 hash of a file in the github repo.
    """
    context.log.debug(f"retrieving contents of {path}")
    file_contents = github_connector_repo.get_contents(path)

    # calculate the md5 hash of the file contents
    context.log.debug(f"calculating md5 hash of {path}")
    md5_hash = hashlib.md5()
    md5_hash.update(file_contents.decoded_content)
    base_64_value = base64.b64encode(md5_hash.digest()).decode("utf8")
    return base_64_value


def _get_content_of_github_file(
    context: OpExecutionContext, github_connector_repo: Repository, path: str
) -> str:
    context.log.debug(f"retrieving contents of {path}")
    return github_connector_repo.get_contents(path)


@asset(required_resource_keys={"github_connectors_directory"}, group_name=GROUP_NAME)
@sentry.instrument_asset_op
def github_connector_folders(context):
    """
    Return a list of all the folders in the github connectors directory.
    """
    github_connectors_directory = context.resources.github_connectors_directory

    folder_names = [
        item.name for item in github_connectors_directory if item.type == "dir"
    ]
    return Output(folder_names, metadata={"preview": folder_names})


@asset(
    required_resource_keys={
        "github_connector_repo",
        "github_connectors_metadata_files",
    },
    group_name=GROUP_NAME,
)
def github_metadata_file_md5s(context):
    """
    Return a list of all the folders in the github connectors directory.
    """
    github_connector_repo = context.resources.github_connector_repo
    github_connectors_metadata_files = (
        context.resources.github_connectors_metadata_files
    )

    metadata_file_paths = {
        metadata_file["path"]: {
            "md5": _get_md5_of_github_file(
                context, github_connector_repo, metadata_file["path"]
            ),
            "last_modified": metadata_file["last_modified"],
        }
        for metadata_file in github_connectors_metadata_files
    }

    return Output(metadata_file_paths, metadata={"preview": metadata_file_paths})


@asset(
    required_resource_keys={
        "github_connector_repo",
        "github_connectors_metadata_files",
    },
    group_name=GROUP_NAME,
)
def github_metadata_definitions(context):
    """
    Return a list of all metadata definitions hosted on our github repo
    """
    github_connector_repo = context.resources.github_connector_repo
    github_connectors_metadata_files = (
        context.resources.github_connectors_metadata_files
    )

    metadata_definitions = []
    for metadata_file in github_connectors_metadata_files:
        metadata_raw = _get_content_of_github_file(
            context, github_connector_repo, metadata_file["path"]
        )
        metadata_dict = yaml.safe_load(metadata_raw.decoded_content)
        if metadata_dict.get("data").get("supportLevel") == "archived":
            print(
                f"Skipping archived connector: {metadata_dict.get('data').get('dockerRepository')}"
            )
            continue
        metadata_definitions.append(
            LatestMetadataEntry(
                metadata_definition=MetadataDefinition.parse_obj(metadata_dict),
                last_modified=metadata_file["last_modified"],
            )
        )

    return Output(
        metadata_definitions,
        metadata={"preview": [md.json() for md in metadata_definitions]},
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING

from consts import LOCAL_BUILD_PLATFORM
from jinja2 import Environment, PackageLoader, select_autoescape
from pipelines import hacks
from pipelines.airbyte_ci.connectors.build_image.steps import run_connector_build
from pipelines.airbyte_ci.connectors.context import ConnectorContext
from pipelines.airbyte_ci.connectors.reports import ConnectorReport
from pipelines.airbyte_ci.steps.base_image import UpdateBaseImageMetadata
from pipelines.airbyte_ci.steps.bump_version import BumpConnectorVersion
from pipelines.airbyte_ci.steps.changelog import AddChangelogEntry
from pipelines.airbyte_ci.steps.pull_request import CreateOrUpdatePullRequest

from .steps import DependencyUpdate, GetDependencyUpdates, PoetryUpdate

if TYPE_CHECKING:
    from typing import Dict, Iterable, List, Set, Tuple

    from anyio import Semaphore
    from github import PullRequest
    from pipelines.models.steps import StepResult

UP_TO_DATE_PR_LABEL = "up-to-date"
AUTO_MERGE_PR_LABEL = "auto-merge"
DEFAULT_PR_LABELS = [UP_TO_DATE_PR_LABEL]
BUMP_TYPE = "patch"
CHANGELOG_ENTRY_COMMENT = "Update dependencies"


## HELPER FUNCTIONS


def get_pr_body(
    context: ConnectorContext,
    step_results: Iterable[StepResult],
    dependency_updates: Iterable[DependencyUpdate],
) -> str:
    env = Environment(
        loader=PackageLoader("pipelines.airbyte_ci.connectors.up_to_date"),
        autoescape=select_autoescape(),
        trim_blocks=False,
        lstrip_blocks=True,
    )
    template = env.get_template("up_to_date_pr_body.md.j2")
    latest_docker_image = f"{context.connector.metadata['dockerRepository']}:latest"
    return template.render(
        connector_technical_name=context.connector.technical_name,
        step_results=step_results,
        dependency_updates=dependency_updates,
        connector_latest_docker_image=latest_docker_image,
    )


def get_pr_creation_arguments(
    modified_files: Iterable[Path],
    context: ConnectorContext,
    step_results: Iterable[StepResult],
    dependency_updates: Iterable[DependencyUpdate],
) -> Tuple[Tuple, Dict]:
    return (
        (modified_files,),
        {
            "branch_id": f"up-to-date/{context.connector.technical_name}",
            "commit_message": "[up-to-date]"  # << We can skip Vercel builds if this is in the commit message
            + "; ".join(
                step_result.step.title
                for step_result in step_results
                if step_result.success
            ),
            "pr_title": f"🐙 {context.connector.technical_name}: run up-to-date pipeline [{datetime.now(timezone.utc).strftime('%Y-%m-%d')}]",
            "pr_body": get_pr_body(context, step_results, dependency_updates),
        },
    )


## MAIN FUNCTION
async def run_connector_up_to_date_pipeline(
    context: ConnectorContext,
    semaphore: Semaphore,
    create_pull_request: bool = False,
    auto_merge: bool = False,
    specific_dependencies: List[str] = [],
    bump_connector_version: bool = True,
) -> ConnectorReport:
    async with semaphore:
        async with context:
            step_results: List[StepResult] = []
            all_modified_files: Set[Path] = set()
            created_pr: PullRequest.PullRequest | None = None
            new_version: str | None = None

            connector_directory = await context.get_connector_dir()
            upgrade_base_image_in_metadata = UpdateBaseImageMetadata(
                context, connector_directory
            )
            upgrade_base_image_in_metadata_result = (
                await upgrade_base_image_in_metadata.run()
            )
            step_results.append(upgrade_base_image_in_metadata_result)
            if upgrade_base_image_in_metadata_result.success:
                connector_directory = upgrade_base_image_in_metadata_result.output[
                    "updated_connector_directory"
                ]
                exported_modified_files = (
                    await upgrade_base_image_in_metadata.export_modified_files(
                        context.connector.code_directory
                    )
                )
                context.logger.info(
                    f"Exported files following the base image upgrade: {exported_modified_files}"
                )
                all_modified_files.update(exported_modified_files)

            if context.connector.is_using_poetry:
                # We run the poetry update step after the base image upgrade because the base image upgrade may change the python environment
                poetry_update = PoetryUpdate(
                    context,
                    connector_directory,
                    specific_dependencies=specific_dependencies,
                )
                poetry_update_result = await poetry_update.run()
                step_results.append(poetry_update_result)
                if poetry_update_result.success:
                    exported_modified_files = await poetry_update.export_modified_files(
                        context.connector.code_directory
                    )
                    context.logger.info(
                        f"Exported files following the Poetry update: {exported_modified_files}"
                    )
                    all_modified_files.update(exported_modified_files)
                    connector_directory = poetry_update_result.output

            one_previous_step_is_successful = any(
                step_result.success for step_result in step_results
            )

            # NOTE:
            # BumpConnectorVersion will already work for manifest-only and Java connectors too
            if bump_connector_version and one_previous_step_is_successful:
                bump_version = BumpConnectorVersion(
                    context, connector_directory, BUMP_TYPE
                )
                bump_version_result = await bump_version.run()
                step_results.append(bump_version_result)
                if bump_version_result.success:
                    new_version = bump_version.new_version
                    exported_modified_files = await bump_version.export_modified_files(
                        context.connector.code_directory
                    )
                    context.logger.info(
                        f"Exported files following the version bump: {exported_modified_files}"
                    )
                    all_modified_files.update(exported_modified_files)

            # Only create the PR if the flag is on, and if there's anything to contribute
            create_pull_request = (
                create_pull_request
                and one_previous_step_is_successful
                and bump_version_result.success
            )

            # We run build and get dependency updates only if we are creating a pull request,
            # to fill the PR body with the correct information about what exactly got updated.
            if create_pull_request:
                # Building connector images is also universal across connector technologies.
                build_result = await run_connector_build(context)
                step_results.append(build_result)
                dependency_updates: List[DependencyUpdate] = []

                if build_result.success:
                    built_connector_container = build_result.output[
                        LOCAL_BUILD_PLATFORM
                    ]

                    # Dependencies here mean Syft deps in the container image itself, not framework-level deps.
                    get_dependency_updates = GetDependencyUpdates(context)
                    dependency_updates_result = await get_dependency_updates.run(
                        built_connector_container
                    )
                    step_results.append(dependency_updates_result)
                    dependency_updates = dependency_updates_result.output

                # We open a PR even if build is failing.
                # This might allow a developer to fix the build in the PR.
                initial_pr_creation = CreateOrUpdatePullRequest(
                    context,
                    labels=DEFAULT_PR_LABELS,
                    # Reduce pressure on rate limit, since we need to push a
                    # a follow-on commit anyway once we have the PR number:
                    skip_ci=True,
                )
                pr_creation_args, pr_creation_kwargs = get_pr_creation_arguments(
                    all_modified_files, context, step_results, dependency_updates
                )
                initial_pr_creation_result = await initial_pr_creation.run(
                    *pr_creation_args, **pr_creation_kwargs
                )
                step_results.append(initial_pr_creation_result)
                if initial_pr_creation_result.success:
                    created_pr = initial_pr_creation_result.output

            if new_version and created_pr:
                documentation_directory = await context.get_repo_dir(
                    include=[
                        str(context.connector.local_connector_documentation_directory)
                    ]
                ).directory(
                    str(context.connector.local_connector_documentation_directory)
                )

                changelog_entry_comment = hacks.determine_changelog_entry_comment(
                    upgrade_base_image_in_metadata_result, CHANGELOG_ENTRY_COMMENT
                )
                add_changelog_entry = AddChangelogEntry(
                    context,
                    documentation_directory,
                    new_version,
                    changelog_entry_comment,
                    created_pr.number,
                )
                add_changelog_entry_result = await add_changelog_entry.run()
                step_results.append(add_changelog_entry_result)
                if add_changelog_entry_result.success:
                    # File path modified by the changelog entry step are relative to the repo root
                    exported_modified_files = (
                        await add_changelog_entry.export_modified_files(
                            context.connector.local_connector_documentation_directory
                        )
                    )
                    context.logger.info(
                        f"Exported files following the changelog entry: {exported_modified_files}"
                    )
                    all_modified_files.update(exported_modified_files)
                    final_labels = (
                        DEFAULT_PR_LABELS + [AUTO_MERGE_PR_LABEL]
                        if auto_merge
                        else DEFAULT_PR_LABELS
                    )
                    post_changelog_pr_update = CreateOrUpdatePullRequest(
                        context,
                        labels=final_labels,
                        # For this 'up-to-date' pipeline, we want GitHub to merge organically
                        # if/when all required checks pass. Maintainers can also easily disable
                        # auto-merge if they want to review or update the PR before merging.
                        github_auto_merge=auto_merge,
                        skip_ci=False,
                    )
                    pr_creation_args, pr_creation_kwargs = get_pr_creation_arguments(
                        all_modified_files, context, step_results, dependency_updates
                    )
                    post_changelog_pr_update_result = (
                        await post_changelog_pr_update.run(
                            *pr_creation_args, **pr_creation_kwargs
                        )
                    )
                    step_results.append(post_changelog_pr_update_result)

            report = ConnectorReport(context, step_results, name="UP-TO-DATE RESULTS")
            context.report = report
    return report



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import datetime

from consts import AMAZONCORRETTO_IMAGE
from dagger import CacheVolume, Container, File, Platform
from pipelines.airbyte_ci.connectors.context import ConnectorContext, PipelineContext
from pipelines.dagger.actions.connector.hooks import finalize_build
from pipelines.dagger.actions.connector.normalization import (
    DESTINATION_NORMALIZATION_BUILD_CONFIGURATION,
    with_normalization,
)
from pipelines.helpers.utils import deprecated, sh_dash_c


@deprecated(
    "This function is deprecated. Please declare an explicit base image to use in the java connector metadata."
)
def with_integration_base(
    context: PipelineContext, build_platform: Platform
) -> Container:
    return (
        context.dagger_client.container(platform=build_platform)
        .from_("amazonlinux:2022.0.20220831.1")
        .with_workdir("/airbyte")
        .with_file(
            "base.sh",
            context.get_repo_dir(
                "airbyte-integrations/bases/base", include=["base.sh"]
            ).file("base.sh"),
        )
        .with_env_variable("AIRBYTE_ENTRYPOINT", "/airbyte/base.sh")
        .with_label("io.airbyte.version", "0.1.0")
        .with_label("io.airbyte.name", "airbyte/integration-base")
    )


@deprecated(
    "This function is deprecated. Please declare an explicit base image to use in the java connector metadata."
)
def with_integration_base_java(
    context: PipelineContext, build_platform: Platform
) -> Container:
    integration_base = with_integration_base(context, build_platform)
    yum_packages_to_install = [
        "tar",  # required to untar java connector binary distributions.
        "openssl",  # required because we need to ssh and scp sometimes.
        "findutils",  # required for xargs, which is shipped as part of findutils.
    ]
    return (
        context.dagger_client.container(platform=build_platform)
        # Use a linux+jdk base image with long-term support, such as amazoncorretto.
        .from_(AMAZONCORRETTO_IMAGE)
        # Bust the cache on a daily basis to get fresh packages.
        .with_env_variable(
            "DAILY_CACHEBUSTER", datetime.datetime.now().strftime("%Y-%m-%d")
        )
        # Install a bunch of packages as early as possible.
        .with_exec(
            sh_dash_c(
                [
                    # Update first, but in the same .with_exec step as the package installation.
                    # Otherwise, we risk caching stale package URLs.
                    "yum update -y --security",
                    #
                    f"yum install -y {' '.join(yum_packages_to_install)}",
                    # Remove any dangly bits.
                    "yum clean all",
                ]
            )
        )
        # Add what files we need to the /airbyte directory.
        # Copy base.sh from the airbyte/integration-base image.
        .with_directory("/airbyte", integration_base.directory("/airbyte"))
        .with_workdir("/airbyte")
        # Download a utility jar from the internet.
        .with_file(
            "dd-java-agent.jar",
            context.dagger_client.http("https://dtdg.co/latest-java-tracer"),
        )
        # Copy javabase.sh from the git repo.
        .with_file(
            "javabase.sh",
            context.get_repo_dir(
                "airbyte-integrations/bases/base-java", include=["javabase.sh"]
            ).file("javabase.sh"),
        )
        # Set a bunch of env variables used by base.sh.
        .with_env_variable("AIRBYTE_SPEC_CMD", "/airbyte/javabase.sh --spec")
        .with_env_variable("AIRBYTE_CHECK_CMD", "/airbyte/javabase.sh --check")
        .with_env_variable("AIRBYTE_DISCOVER_CMD", "/airbyte/javabase.sh --discover")
        .with_env_variable("AIRBYTE_READ_CMD", "/airbyte/javabase.sh --read")
        .with_env_variable("AIRBYTE_WRITE_CMD", "/airbyte/javabase.sh --write")
        .with_env_variable("AIRBYTE_ENTRYPOINT", "/airbyte/base.sh")
        # Set image labels.
        .with_label("io.airbyte.version", "0.1.2")
        .with_label("io.airbyte.name", "airbyte/integration-base-java")
    )


@deprecated(
    "This function is deprecated. Please declare an explicit base image to use in the java connector metadata."
)
def with_integration_base_java_and_normalization(
    context: ConnectorContext, build_platform: Platform
) -> Container:
    yum_packages_to_install = [
        "python3",
        "python3-devel",
        "jq",
        "sshpass",
        "git",
    ]

    additional_yum_packages = DESTINATION_NORMALIZATION_BUILD_CONFIGURATION[
        context.connector.technical_name
    ]["yum_packages"]
    yum_packages_to_install += additional_yum_packages

    dbt_adapter_package = DESTINATION_NORMALIZATION_BUILD_CONFIGURATION[
        context.connector.technical_name
    ]["dbt_adapter"]
    assert isinstance(dbt_adapter_package, str)
    normalization_integration_name = DESTINATION_NORMALIZATION_BUILD_CONFIGURATION[
        context.connector.technical_name
    ]["integration_name"]
    assert isinstance(normalization_integration_name, str)

    pip_cache: CacheVolume = context.dagger_client.cache_volume("pip_cache")

    return (
        with_integration_base_java(context, build_platform)
        # Bust the cache on a daily basis to get fresh packages.
        .with_env_variable(
            "DAILY_CACHEBUSTER", datetime.datetime.now().strftime("%Y-%m-%d")
        )
        .with_exec(
            sh_dash_c(
                [
                    "yum update -y --security",
                    f"yum install -y {' '.join(yum_packages_to_install)}",
                    "yum clean all",
                    "alternatives --install /usr/bin/python python /usr/bin/python3 60",
                ]
            )
        )
        .with_mounted_cache("/root/.cache/pip", pip_cache)
        .with_exec(
            sh_dash_c(
                [
                    "python -m ensurepip --upgrade",
                    # Workaround for https://github.com/yaml/pyyaml/issues/601
                    "pip3 install 'Cython<3.0' 'pyyaml~=5.4' --no-build-isolation",
                    # Required for dbt https://github.com/dbt-labs/dbt-core/issues/7075
                    "pip3 install 'pytz~=2023.3'",
                    f"pip3 install {dbt_adapter_package}",
                    # amazon linux 2 isn't compatible with urllib3 2.x, so force 1.x
                    "pip3 install 'urllib3<2'",
                ]
            )
        )
        .with_directory(
            "airbyte_normalization",
            with_normalization(context, build_platform).directory("/airbyte"),
        )
        .with_workdir("airbyte_normalization")
        .with_exec(sh_dash_c(["mv * .."]))
        .with_workdir("/airbyte")
        .with_exec(["rm", "-rf", "airbyte_normalization"], use_entrypoint=True)
        .with_workdir("/airbyte/normalization_code")
        .with_exec(["pip3", "install", "."], use_entrypoint=True)
        .with_workdir("/airbyte/normalization_code/dbt-template/")
        .with_exec(["dbt", "deps"], use_entrypoint=True)
        .with_workdir("/airbyte")
        .with_file(
            "run_with_normalization.sh",
            context.get_repo_dir(
                "airbyte-integrations/bases/base-java",
                include=["run_with_normalization.sh"],
            ).file("run_with_normalization.sh"),
        )
        .with_env_variable(
            "AIRBYTE_NORMALIZATION_INTEGRATION", normalization_integration_name
        )
        .with_env_variable("AIRBYTE_ENTRYPOINT", "/airbyte/run_with_normalization.sh")
    )


async def with_airbyte_java_connector(
    context: ConnectorContext, connector_java_tar_file: File, build_platform: Platform
) -> Container:
    application = context.connector.technical_name

    build_stage = (
        with_integration_base_java(context, build_platform)
        .with_workdir("/airbyte")
        .with_env_variable("APPLICATION", context.connector.technical_name)
        .with_file(f"{application}.tar", connector_java_tar_file)
        .with_exec(
            sh_dash_c(
                [
                    f"tar xf {application}.tar --strip-components=1",
                    f"rm -rf {application}.tar",
                ]
            )
        )
    )
    # TODO: remove the condition below once all connectors have a base image declared in their metadata.
    if (
        "connectorBuildOptions" in context.connector.metadata
        and "baseImage" in context.connector.metadata["connectorBuildOptions"]
    ):
        base_image_address = context.connector.metadata["connectorBuildOptions"][
            "baseImage"
        ]
        context.logger.info(
            f"Using base image {base_image_address} from connector metadata to build connector."
        )
        base = context.dagger_client.container(platform=build_platform).from_(
            base_image_address
        )
    elif (
        context.connector.supports_normalization
        and DESTINATION_NORMALIZATION_BUILD_CONFIGURATION[
            context.connector.technical_name
        ]["supports_in_connector_normalization"]
    ):
        context.logger.warn(
            f"Connector {context.connector.technical_name} has in-connector normalization enabled. This is supposed to be deprecated. "
            f"Please declare a base image address in the connector metadata.yaml file (connectorBuildOptions.baseImage)."
        )
        base = with_integration_base_java_and_normalization(
            context, build_platform
        ).with_entrypoint(["/airbyte/run_with_normalization.sh"])
    else:
        context.logger.warn(
            f"Connector {context.connector.technical_name} does not declare a base image in its connector metadata. "
            f"Please declare a base image address in the connector metadata.yaml file (connectorBuildOptions.baseImage)."
        )
        base = with_integration_base_java(context, build_platform).with_entrypoint(
            ["/airbyte/base.sh"]
        )

    current_user = (await base.with_exec(["whoami"]).stdout()).strip()
    connector_container = (
        base.with_workdir("/airbyte")
        .with_env_variable("APPLICATION", application)
        .with_mounted_directory(
            "built_artifacts", build_stage.directory("/airbyte"), owner=current_user
        )
        .with_exec(sh_dash_c(["mv built_artifacts/* ."]))
    )
    return await finalize_build(context, connector_container)



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Registry rebuild operations.

This module provides the core logic for rebuilding the entire connector registry
by reading all connector metadata from a source GCS bucket and writing it to
a configurable output target (local directory, GCS bucket, or S3 bucket)
using fsspec for unified filesystem access.
"""

from __future__ import annotations

import json
import logging
import os
import subprocess
import sys
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal

import fsspec
import gcsfs
import s3fs

from airbyte_ops_mcp.registry._constants import (
    METADATA_FOLDER,
    PROD_METADATA_SERVICE_BUCKET_NAME,
)
from airbyte_ops_mcp.registry._gcs_helpers import get_gcs_credentials_token

logger = logging.getLogger(__name__)

OutputMode = Literal["local", "gcs", "s3"]

_GCLOUD_CP_MAX_WORKERS = 80
"""Maximum number of parallel `gcloud storage cp` subprocesses."""


@dataclass
class RebuildResult:
    """Result of a registry rebuild operation."""

    source_bucket: str
    output_mode: OutputMode
    output_root: str
    connectors_processed: int = 0
    blobs_copied: int = 0
    blobs_skipped: int = 0
    errors: list[str] = field(default_factory=list)
    dry_run: bool = False

    @property
    def status(self) -> str:
        """Return the status of the rebuild operation."""
        if self.dry_run:
            return "dry-run"
        if self.errors:
            return "completed-with-errors"
        return "success"

    def summary(self) -> str:
        """Return a human-readable summary."""
        return (
            f"[{self.status}] Rebuilt {self.connectors_processed} connectors, "
            f"{self.blobs_copied} blobs copied, {self.blobs_skipped} skipped, "
            f"{len(self.errors)} errors. Output: {self.output_root}"
        )


def _validate_not_prod_bucket(bucket_name: str) -> None:
    """Raise ValueError if the bucket is the prod bucket."""
    if bucket_name == PROD_METADATA_SERVICE_BUCKET_NAME:
        raise ValueError(
            f"Writing to the production bucket '{PROD_METADATA_SERVICE_BUCKET_NAME}' "
            "is categorically disallowed by create-mirror."
        )


def _make_source_fs() -> tuple[gcsfs.GCSFileSystem, dict[str, str] | None]:
    """Create a gcsfs filesystem for reading from GCS.

    Returns:
        Tuple of (filesystem, token) for reuse on the output side if needed.
    """
    token = get_gcs_credentials_token()
    fs = gcsfs.GCSFileSystem(token=token)
    return fs, token


def _make_output_fs(
    output_mode: OutputMode,
    output_root: str,
    gcs_bucket: str | None = None,
    s3_bucket: str | None = None,
    gcs_token: dict[str, str] | None = None,
) -> tuple[fsspec.AbstractFileSystem, str]:
    """Create an fsspec filesystem for writing to the output target.

    Args:
        output_mode: "local", "gcs", or "s3".
        output_root: The resolved output root path (local dir or prefix).
        gcs_bucket: Target GCS bucket (required for gcs mode).
        s3_bucket: Target S3 bucket (required for s3 mode).
        gcs_token: GCS credentials token dict (reused from source if available).

    Returns:
        Tuple of (filesystem, base_path) for the output target.
    """
    if output_mode == "local":
        fs = fsspec.filesystem("file")
        return fs, output_root

    if output_mode == "gcs":
        if not gcs_bucket:
            raise ValueError("gcs_bucket is required for GCS output mode.")
        _validate_not_prod_bucket(gcs_bucket)
        token = gcs_token or get_gcs_credentials_token()
        fs = gcsfs.GCSFileSystem(token=token)
        prefix = f"{output_root.rstrip('/')}/" if output_root else ""
        return fs, f"{gcs_bucket}/{prefix}"

    # s3 mode
    if not s3_bucket:
        raise ValueError("s3_bucket is required for S3 output mode.")
    fs = s3fs.S3FileSystem()
    prefix = f"{output_root.rstrip('/')}/" if output_root else ""
    return fs, f"{s3_bucket}/{prefix}"


def _resolve_local_output_root(output_path_root: str | None) -> str:
    """Resolve the local output root path.

    If no path is provided, creates a new temp directory.
    """
    if output_path_root:
        path = Path(output_path_root)
        path.mkdir(parents=True, exist_ok=True)
        return str(path)

    return tempfile.mkdtemp(prefix="registry-rebuild-")


def rebuild_registry(
    source_bucket: str,
    output_mode: OutputMode,
    output_path_root: str | None = None,
    gcs_bucket: str | None = None,
    s3_bucket: str | None = None,
    dry_run: bool = False,
    connector_name: list[str] | None = None,
) -> RebuildResult:
    """Rebuild the entire registry from a source GCS bucket to an output target.

    Reads all connector metadata blobs from the source GCS bucket and copies
    them to the output target using fsspec for unified filesystem access.

    The output targets are:
    - local: Write to a local directory tree.
    - gcs: Copy to a GCS bucket (must not be the prod bucket).
    - s3: Copy to an S3 bucket.

    Args:
        source_bucket: The GCS bucket to read from (typically prod).
        output_mode: Where to write: "local", "gcs", or "s3".
        output_path_root: Root path/prefix for output. For local mode, if None
            creates a temp directory. For GCS/S3, prepended to all blob paths.
        gcs_bucket: Target GCS bucket name (required if output_mode="gcs").
        s3_bucket: Target S3 bucket name (required if output_mode="s3").
        dry_run: If True, report what would be done without writing.
        connector_name: If provided, only rebuild these connector names
            (e.g. ["source-faker", "destination-bigquery"]). If None, rebuilds all.

    Returns:
        RebuildResult with details of the operation.

    Raises:
        ValueError: If target is the prod bucket, or required bucket arg is missing.
    """
    if output_mode == "gcs" and gcs_bucket:
        _validate_not_prod_bucket(gcs_bucket)

    # Resolve output root for local mode
    effective_output_root = output_path_root or ""
    if output_mode == "local":
        effective_output_root = _resolve_local_output_root(output_path_root)

    result = RebuildResult(
        source_bucket=source_bucket,
        output_mode=output_mode,
        output_root=effective_output_root,
        dry_run=dry_run,
    )

    # Create source filesystem (always GCS)
    source_fs, gcs_token = _make_source_fs()
    source_base = f"{source_bucket}/{METADATA_FOLDER}"

    # List files under metadata/ in the source bucket
    if connector_name:
        _log_progress(
            "Listing blobs for %d connectors under gs://%s/...",
            len(connector_name),
            source_base,
        )
        source_paths: list[str] = []
        for name in connector_name:
            connector_prefix = f"{source_base}/airbyte/{name}"
            found = source_fs.find(connector_prefix)
            source_paths.extend(found)
            _log_progress("  %s: %d blobs", name, len(found))
    else:
        _log_progress("Listing all blobs under gs://%s/...", source_base)
        source_paths = source_fs.find(source_base)

    if not source_paths:
        _log_progress("No blobs found under gs://%s/", source_base)
        return result

    total_blobs = len(source_paths)
    _log_progress("Found %d blobs to process", total_blobs)

    # Create output filesystem
    output_fs, output_base = _make_output_fs(
        output_mode=output_mode,
        output_root=effective_output_root,
        gcs_bucket=gcs_bucket,
        s3_bucket=s3_bucket,
        gcs_token=gcs_token,
    )

    # Collect connector names and compute relative paths for all blobs.
    bucket_prefix = f"{source_bucket}/"
    blob_relative_paths: list[str] = []
    connector_names: set[str] = set()

    for source_path in source_paths:
        relative_path = source_path
        if source_path.startswith(bucket_prefix):
            relative_path = source_path[len(bucket_prefix) :]
        blob_relative_paths.append(relative_path)

        parts = relative_path.split("/")
        if len(parts) >= 3:
            connector_names.add(parts[2])

    if dry_run:
        result.blobs_copied = total_blobs
        result.connectors_processed = len(connector_names)
        _log_progress(
            "[DRY RUN] Would copy %d blobs (%d connectors)",
            total_blobs,
            len(connector_names),
        )
        _log_progress(result.summary())
        return result

    # Use GCS-native server-side copy for GCS→GCS mirrors.
    if output_mode == "gcs" and gcs_bucket:
        result = _gcs_native_copy(
            source_bucket=source_bucket,
            dest_bucket_name=gcs_bucket,
            dest_prefix=effective_output_root,
            blob_relative_paths=blob_relative_paths,
            connector_names=connector_names,
            gcs_token=gcs_token,
            result=result,
            connector_name_filter=connector_name,
        )
        return result

    # Fallback: fsspec-based copy for local and S3 output modes.
    result = _fsspec_copy(
        source_fs=source_fs,
        source_paths=source_paths,
        output_fs=output_fs,
        output_base=output_base,
        output_mode=output_mode,
        source_bucket=source_bucket,
        blob_relative_paths=blob_relative_paths,
        connector_names=connector_names,
        result=result,
    )
    return result


def _gcs_native_copy(
    source_bucket: str,
    dest_bucket_name: str,
    dest_prefix: str,
    blob_relative_paths: list[str],
    connector_names: set[str],
    gcs_token: dict[str, str] | None,
    result: RebuildResult,
    connector_name_filter: list[str] | None = None,
) -> RebuildResult:
    """Copy blobs using GCS-native server-side copy (no data through Python).

    Delegates to `gcloud storage cp` which uses built-in parallelism
    and server-side copy (no data transits through this machine) for
    same-provider GCS-to-GCS transfers.  This is significantly faster
    than issuing individual `copy_blob` API calls from the Python SDK.
    """
    token = gcs_token or get_gcs_credentials_token()
    total_blobs = len(blob_relative_paths)
    prefix_part = f"{dest_prefix.rstrip('/')}/" if dest_prefix else ""

    # Build (source, dest) pairs.
    # Per-connector pairs target metadata/airbyte/{name}/ on both sides.
    # The whole-registry pair targets the parent prefix because --recursive
    # preserves the source directory name (avoids double-nesting metadata/).
    pairs = (
        [
            (
                f"gs://{source_bucket}/{METADATA_FOLDER}/airbyte/{name}/",
                f"gs://{dest_bucket_name}/{prefix_part}{METADATA_FOLDER}/airbyte/{name}/",
            )
            for name in connector_name_filter
        ]
        if connector_name_filter
        else [
            (
                f"gs://{source_bucket}/{METADATA_FOLDER}/",
                f"gs://{dest_bucket_name}/{prefix_part}",
            )
        ]
    )

    _log_progress(
        "Starting gcloud storage copy: %d blobs -> gs://%s/%s",
        total_blobs,
        dest_bucket_name,
        prefix_part,
    )

    def _run_one_copy(
        pair: tuple[str, str],
        env: dict[str, str],
    ) -> str | None:
        """Run a single gcloud storage cp and return an error string or None."""
        src, dst = pair
        _log_progress("  Copying %s -> %s ...", src, dst)
        cmd = [
            "gcloud",
            "storage",
            "cp",
            "--recursive",
            "--continue-on-error",
            src,
            dst,
        ]
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            env=env,
            stdin=subprocess.DEVNULL,
        )
        line_count = 0
        tail_buffer: list[str] = []  # keep only last 10 lines for errors
        assert proc.stdout is not None  # for type checkers
        for line in proc.stdout:
            stripped = line.rstrip()
            if stripped:
                line_count += 1
                tail_buffer.append(stripped)
                if len(tail_buffer) > 10:
                    tail_buffer.pop(0)
                if line_count % 500 == 0:
                    _log_progress(
                        "  gcloud progress (%s): %d operations...",
                        src.split("/")[-2],
                        line_count,
                    )
        rc = proc.wait()
        if rc != 0:
            tail = "\n".join(tail_buffer) if tail_buffer else "(no output)"
            return f"gcloud storage cp failed for {src} (exit {rc}):\n{tail}"
        return None

    # If we're using explicit service-account creds, write them to a temp file
    # and point gcloud at it via CLOUDSDK_AUTH_CREDENTIAL_FILE_OVERRIDE.
    #
    # If we're using ADC (token is None), do NOT set the override; let gcloud
    # resolve credentials via its normal ADC mechanisms.
    creds_path: str | None = None
    gcloud_env = dict(os.environ)
    if token is not None:
        with tempfile.NamedTemporaryFile(
            mode="w",
            suffix=".json",
            delete=False,
        ) as creds_fd:
            json.dump(token, creds_fd)
            creds_path = creds_fd.name
        gcloud_env["CLOUDSDK_AUTH_CREDENTIAL_FILE_OVERRIDE"] = creds_path

    t_start = time.monotonic()

    try:
        # Run all copy operations in parallel.
        with ThreadPoolExecutor(
            max_workers=min(len(pairs), _GCLOUD_CP_MAX_WORKERS)
        ) as executor:
            futures = {executor.submit(_run_one_copy, p, gcloud_env): p for p in pairs}
            for future in as_completed(futures):
                error = future.result()
                if error:
                    result.errors.append(error)
                    _log_progress(error)
    finally:
        if creds_path:
            os.unlink(creds_path)

    elapsed = time.monotonic() - t_start

    # Each pair covers a connector prefix; estimate blobs per pair
    # proportionally so partial failures don't zero out the count.
    failed_pairs = len(result.errors)
    successful_pairs = len(pairs) - failed_pairs
    if len(pairs) > 0:
        result.blobs_copied = total_blobs * successful_pairs // len(pairs)
    else:
        result.blobs_copied = 0
    result.connectors_processed = len(connector_names)
    _log_progress(
        "%s [%.1fs elapsed, ~%.0f blobs/s]",
        result.summary(),
        elapsed,
        result.blobs_copied / elapsed if elapsed > 0 else 0,
    )

    return result


def _fsspec_copy(
    source_fs: gcsfs.GCSFileSystem,
    source_paths: list[str],
    output_fs: fsspec.AbstractFileSystem,
    output_base: str,
    output_mode: OutputMode,
    source_bucket: str,
    blob_relative_paths: list[str],
    connector_names: set[str],
    result: RebuildResult,
) -> RebuildResult:
    """Fallback copy using fsspec (read-through-Python) for non-GCS targets."""
    logger.info(
        "Starting fsspec copy from bucket %s to %s (%d blobs)",
        source_bucket,
        output_base,
        len(source_paths),
    )
    total_blobs = len(source_paths)

    for idx, (source_path, relative_path) in enumerate(
        zip(source_paths, blob_relative_paths)
    ):
        if output_mode == "local":
            output_path = f"{output_base}/{relative_path}"
        else:
            output_path = f"{output_base}{relative_path}"

        if output_mode == "local":
            Path(output_path).parent.mkdir(parents=True, exist_ok=True)

        try:
            with source_fs.open(source_path, "rb") as src:
                content = src.read()
            with output_fs.open(output_path, "wb") as dst:
                dst.write(content)
            result.blobs_copied += 1
        except Exception as exc:
            result.errors.append(f"Failed to copy {source_path}: {exc}")
            logger.warning("Failed to copy %s: %s", source_path, exc)

        if (idx + 1) % 1000 == 0 or idx == 0:
            _log_progress(
                "Progress: %d/%d blobs copied (%d connectors)",
                result.blobs_copied,
                total_blobs,
                len(connector_names),
            )

    result.connectors_processed = len(connector_names)
    _log_progress(result.summary())
    return result


def _log_progress(msg: str, *args: object) -> None:
    """Log a progress message to both the logger and stderr."""
    logger.info(msg, *args)
    formatted = msg % args if args else msg
    print(formatted, file=sys.stderr, flush=True)



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from pathlib import Path
from typing import ClassVar, Iterable

from pydash.objects import get  # type: ignore

from airbyte_ops_mcp.connector_ops.utils import (
    Connector,
    ConnectorLanguage,
)  # type: ignore
from airbyte_ops_mcp.connector_qa import consts
from airbyte_ops_mcp.connector_qa.models import (
    Check,
    CheckCategory,
    CheckResult,
)


class SecurityCheck(Check):
    category = CheckCategory.SECURITY


class CheckConnectorUsesHTTPSOnly(SecurityCheck):
    name = "Connectors must use HTTPS only"
    description = (
        "Connectors must use HTTPS only when making requests to external services."
    )
    requires_metadata = False
    runs_on_released_connectors = False

    ignore_comment = "# ignore-https-check"  # Define the ignore comment pattern

    ignored_directories_for_https_checks: ClassVar[set[str]] = {
        ".venv",
        "tests",
        "unit_tests",
        "integration_tests",
        "test-integration",
        "build",
        "source-file",
        ".pytest_cache",
        "acceptance_tests_logs",
        ".hypothesis",
        ".ruff_cache",
        "htmlcov",
    }

    ignored_file_name_pattern_for_https_checks: ClassVar[set[str]] = {
        "*Test.java",
        "*.jar",
        "*.pyc",
        "*.gz",
        "*.svg",
        "expected_records.jsonl",
        "expected_records.json",
    }

    ignored_url_prefixes: ClassVar[set[str]] = {
        "http://json-schema.org",
        "http://localhost",
    }

    @staticmethod
    def _read_all_files_in_directory(
        directory: Path,
        ignored_directories: set[str] | None = None,
        ignored_filename_patterns: set[str] | None = None,
    ) -> Iterable[tuple[Path, str]]:
        ignored_directories = (
            ignored_directories if ignored_directories is not None else set()
        )
        ignored_filename_patterns = (
            ignored_filename_patterns
            if ignored_filename_patterns is not None
            else set()
        )

        for path in directory.rglob("*"):
            ignore_directory = any(
                ignored_directory in path.parts
                for ignored_directory in ignored_directories
            )
            ignore_filename = any(
                path.match(ignored_filename_pattern)
                for ignored_filename_pattern in ignored_filename_patterns
            )
            ignore = ignore_directory or ignore_filename
            if path.is_file() and not ignore:
                try:
                    for line in path.read_text().splitlines():
                        yield path, line
                except UnicodeDecodeError:
                    continue

    @staticmethod
    def _line_is_comment(line: str, file_path: Path) -> bool:
        language_comments = {
            ".py": "#",
            ".yml": "#",
            ".yaml": "#",
            ".java": "//",
            ".md": "<!--",
        }

        denote_comment = language_comments.get(file_path.suffix)
        if not denote_comment:
            return False

        trimmed_line = line.lstrip()
        return trimmed_line.startswith(denote_comment)

    def _run(self, connector: Connector) -> CheckResult:
        files_with_http_url = set()

        for filename, line in self._read_all_files_in_directory(
            connector.code_directory,
            self.ignored_directories_for_https_checks,
            self.ignored_file_name_pattern_for_https_checks,
        ):
            line = line.lower()
            if self._line_is_comment(line, filename):
                continue
            if self.ignore_comment in line:
                continue
            for prefix in self.ignored_url_prefixes:
                line = line.replace(prefix, "")
            if "http://" in line:
                files_with_http_url.add(str(filename))

        if files_with_http_url:
            files_with_http_url_message = "\n\t- ".join(files_with_http_url)
            return self.fail(
                connector=connector,
                message=f"The following files have http:// URLs:\n\t- {files_with_http_url_message}",
            )
        return self.pass_(
            connector=connector, message="No file with http:// URLs found"
        )


class CheckConnectorUsesPythonBaseImage(SecurityCheck):
    name = f"Python connectors must not use a {consts.DOCKERFILE_NAME} and must declare their base image in {consts.METADATA_FILE_NAME} file"
    description = f"Connectors must use our Python connector base image (`{consts.AIRBYTE_PYTHON_CONNECTOR_BASE_IMAGE_NAME}`), declared through the `connectorBuildOptions.baseImage` in their `{consts.METADATA_FILE_NAME}`.\nThis is to ensure that all connectors use a base image which is maintained and has security updates."
    applies_to_connector_languages: ClassVar[list] = [
        ConnectorLanguage.PYTHON,
        ConnectorLanguage.LOW_CODE,
        ConnectorLanguage.MANIFEST_ONLY,
    ]

    def _run(self, connector: Connector) -> CheckResult:
        dockerfile_path = connector.code_directory / consts.DOCKERFILE_NAME
        if dockerfile_path.exists():
            return self.create_check_result(
                connector=connector,
                passed=False,
                message=f"{consts.DOCKERFILE_NAME} file exists. Please remove it and declare the base image in {consts.METADATA_FILE_NAME} file with the `connectorBuildOptions.baseImage` key",
            )

        if not get(connector.metadata, "connectorBuildOptions.baseImage"):
            return self.create_check_result(
                connector=connector,
                passed=False,
                message=f"connectorBuildOptions.baseImage key is missing in {consts.METADATA_FILE_NAME} file",
            )
        return self.create_check_result(
            connector=connector,
            passed=True,
            message="Connector uses the Python connector base image",
        )


ENABLED_CHECKS = [CheckConnectorUsesHTTPSOnly(), CheckConnectorUsesPythonBaseImage()]



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Publish locally generated connector version artifacts to a GCS registry bucket.

This module uses `gcsfs.GCSFileSystem` to upload local artifacts
to the appropriate versioned path inside a GCS registry bucket.

The GCS destination path is:

    gs://<bucket>/[<prefix>/]metadata/airbyte/<connector>/<version>/
"""

from __future__ import annotations

import logging
import sys
from dataclasses import dataclass, field
from pathlib import Path

import gcsfs
import yaml

from airbyte_ops_mcp.registry._constants import (
    CONNECTOR_DEPENDENCY_FILE_NAME,
    SBOM_FILE_NAME,
)
from airbyte_ops_mcp.registry._gcs_helpers import get_gcs_credentials_token
from airbyte_ops_mcp.registry._resolve_gcs_paths import (
    dependencies_blob_path,
    sbom_blob_path,
    versioned_blob_root,
)
from airbyte_ops_mcp.registry._sbom_generation import upload_sbom
from airbyte_ops_mcp.registry.store import RegistryStore
from airbyte_ops_mcp.registry.validate import (
    validate_metadata,
)

logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Result model
# ---------------------------------------------------------------------------


@dataclass
class PublishArtifactsResult:
    """Result of a version-artifacts publish operation."""

    connector_name: str
    version: str
    target: str
    gcs_destination: str
    files_uploaded: list[str] = field(default_factory=list)
    errors: list[str] = field(default_factory=list)
    validation_errors: list[str] = field(default_factory=list)
    dry_run: bool = False

    @property
    def success(self) -> bool:
        return len(self.errors) == 0 and len(self.validation_errors) == 0

    @property
    def status(self) -> str:
        if self.dry_run:
            return "dry-run"
        if self.errors or self.validation_errors:
            return "completed-with-errors"
        return "success"


# ---------------------------------------------------------------------------
# Core publish logic
# ---------------------------------------------------------------------------


def _log_progress(msg: str, *args: object) -> None:
    """Log a progress message to both the logger and stderr."""
    logger.info(msg, *args)
    formatted = msg % args if args else msg
    print(formatted, file=sys.stderr, flush=True)


def _check_connector_name_matches_docker_repo(
    connector_name: str,
    artifacts_dir: Path,
) -> str | None:
    """Verify that `connector_name` matches the `dockerRepository` in metadata.

    Returns `None` when the names match (or when metadata is unavailable).
    Returns an error message string when a mismatch is detected.
    """
    metadata_file = artifacts_dir / "metadata.yaml"
    if not metadata_file.is_file():
        return None

    raw_metadata = yaml.safe_load(metadata_file.read_text())
    docker_repo: str = (raw_metadata or {}).get("data", {}).get("dockerRepository", "")
    if not docker_repo:
        return None

    docker_name = docker_repo.removeprefix("airbyte/")
    if docker_name == connector_name:
        return None

    return (
        f"Connector directory name '{connector_name}' does not match "
        f"dockerRepository '{docker_repo}' (expected directory name "
        f"'{docker_name}'). Rename the connector directory to '{docker_name}' "
        f"or update dockerRepository in metadata.yaml to "
        f"'airbyte/{connector_name}'."
    )


def publish_version_artifacts(
    connector_name: str,
    version: str,
    artifacts_dir: Path,
    store: RegistryStore,
    dry_run: bool = False,
    with_validate: bool = True,
) -> PublishArtifactsResult:
    """Publish locally generated artifacts to a GCS registry bucket.

    Uses `gcsfs.GCSFileSystem` to upload the local *artifacts_dir* to the
    versioned path inside the target GCS bucket.

    The target GCS path is:
        `gs://<bucket>/[<prefix>/]metadata/airbyte/<connector>/<version>/`

    Before uploading, this function validates that the `connector_name`
    (derived from the connector directory) matches the `dockerRepository`
    declared in `metadata.yaml`.  A mismatch would cause the registry
    compile step to see duplicate definition-ID entries and fail.

    Args:
        connector_name: Connector name (e.g. `source-faker`).
        version: Version string (e.g. `6.2.38`).
        artifacts_dir: Local directory containing artifacts from `generate`.
        store: Parsed store target containing bucket, prefix, and stage info.
        dry_run: If `True`, report what would be uploaded without writing.
        with_validate: If `True` (default), validate metadata before uploading.
            Pass `False` (`--no-validate`) to skip.

    Returns:
        A `PublishArtifactsResult` describing what was published.

    Raises:
        ValueError: If the connector directory name does not match
            `dockerRepository` in the generated metadata.
    """
    if not artifacts_dir.is_dir():
        raise FileNotFoundError(f"Artifacts directory not found: {artifacts_dir}")

    # Fail fast if the connector directory name doesn't match dockerRepository.
    # A mismatch would publish artifacts under the wrong GCS path and corrupt
    # the registry (duplicate definition-IDs under different directory names).
    mismatch_error = _check_connector_name_matches_docker_repo(
        connector_name, artifacts_dir
    )
    if mismatch_error:
        raise ValueError(mismatch_error)

    # Build the GCS destination path
    bucket_name = store.bucket
    prefix = store.prefix
    blob_root = versioned_blob_root(
        connector_name=connector_name, version=version, store=store
    )
    versioned_dest = f"gcs://{bucket_name}/{blob_root}"

    target_label = f"{bucket_name}/{prefix}" if prefix else bucket_name
    result = PublishArtifactsResult(
        connector_name=connector_name,
        version=version,
        target=target_label,
        gcs_destination=versioned_dest,
        dry_run=dry_run,
    )

    # --- Pre-publish validation ---
    if with_validate:
        metadata_file = artifacts_dir / "metadata.yaml"
        if metadata_file.is_file():
            raw_metadata = yaml.safe_load(metadata_file.read_text())
            metadata_data = (raw_metadata or {}).get("data", {})
            validation = validate_metadata(metadata_data=metadata_data)
            if not validation.passed:
                for err in validation.errors:
                    logger.error("Pre-publish validation error: %s", err)
                result.validation_errors = validation.errors
                return result
            logger.info(
                "Pre-publish validation passed (%d validators).",
                validation.validators_run,
            )
        else:
            logger.warning("No metadata.yaml in artifacts dir; skipping validation.")

    # Enumerate local files
    local_files = sorted(f for f in artifacts_dir.rglob("*") if f.is_file())
    if not local_files:
        result.errors.append(f"No files found in {artifacts_dir}.")
        return result

    _log_progress(
        "Publishing %d artifacts for %s@%s → %s",
        len(local_files),
        connector_name,
        version,
        versioned_dest,
    )

    # Build references used by both dry-run and real upload paths
    deps_file = artifacts_dir / CONNECTOR_DEPENDENCY_FILE_NAME
    has_deps = deps_file.is_file()
    deps_gcs_key = dependencies_blob_path(
        connector_name=connector_name, version=version, store=store
    )

    sbom_file = artifacts_dir / SBOM_FILE_NAME
    has_sbom = sbom_file.is_file()

    if dry_run:
        for f in local_files:
            rel = f.relative_to(artifacts_dir)
            result.files_uploaded.append(str(rel))
            _log_progress("  [DRY RUN] would upload: %s", rel)
        # Report the dual-load of dependencies.json to connector_dependencies/
        if has_deps:
            result.files_uploaded.append(deps_gcs_key)
            _log_progress(
                "  [DRY RUN] would also dual-load: %s → gs://%s/%s",
                CONNECTOR_DEPENDENCY_FILE_NAME,
                bucket_name,
                deps_gcs_key,
            )
        # Report the separate sbom/ upload
        if has_sbom:
            sbom_gcs_key = sbom_blob_path(
                connector_name=connector_name,
                version=version,
                store=store,
            )
            result.files_uploaded.append(sbom_gcs_key)
            _log_progress(
                "  [DRY RUN] would also upload: %s → gs://%s/%s",
                SBOM_FILE_NAME,
                bucket_name,
                sbom_gcs_key,
            )
        return result

    # Authenticate
    token = get_gcs_credentials_token()
    fs = gcsfs.GCSFileSystem(token=token)

    # Strip gcs:// prefix for gcsfs path
    dest_path = versioned_dest.replace("gcs://", "")

    # Upload all files to the versioned path
    _log_progress("Uploading to: %s", versioned_dest)
    for f in local_files:
        rel = f.relative_to(artifacts_dir)
        remote_path = f"{dest_path}/{rel}"
        fs.put(str(f), remote_path)
        result.files_uploaded.append(str(rel))
        _log_progress("  Uploaded: %s", rel)

    # Delete remote files that don't exist locally (sync semantics)
    try:
        remote_files = fs.ls(dest_path, detail=False)
        local_rel_paths = {str(f.relative_to(artifacts_dir)) for f in local_files}
        for remote_file in remote_files:
            # Skip the directory entry itself if it appears in the listing
            if remote_file == dest_path:
                continue
            # Derive the remote relative path, matching upload semantics
            if remote_file.startswith(dest_path + "/"):
                remote_rel = remote_file[len(dest_path) + 1 :]
            else:
                remote_rel = remote_file.split("/")[-1]
            if remote_rel not in local_rel_paths:
                fs.rm(remote_file)
                _log_progress("  Deleted stale remote file: %s", remote_rel)
    except FileNotFoundError:
        pass  # Destination doesn't exist yet, nothing to clean

    _log_progress("Uploaded %d files to %s", len(local_files), versioned_dest)

    # --- Dual-load dependencies.json to the connector_dependencies/ path ---
    if not has_deps:
        logger.debug(
            "No %s in artifacts dir — skipping dual-load.",
            CONNECTOR_DEPENDENCY_FILE_NAME,
        )
    else:
        deps_remote = f"{bucket_name}/{deps_gcs_key}"
        _log_progress(
            "Dual-loading %s to gs://%s",
            CONNECTOR_DEPENDENCY_FILE_NAME,
            deps_remote,
        )
        fs.put(str(deps_file), deps_remote)
        result.files_uploaded.append(deps_gcs_key)
        _log_progress("  Uploaded %s (dual-load)", CONNECTOR_DEPENDENCY_FILE_NAME)

    # --- Upload SBOM to the dedicated sbom/ path in GCS ---
    if not has_sbom:
        logger.debug(
            "No %s in artifacts dir — skipping SBOM dual-load.",
            SBOM_FILE_NAME,
        )
    else:
        sbom_gcs_uri = upload_sbom(
            sbom_path=sbom_file,
            connector_name=connector_name,
            version=version,
            store=store,
            dry_run=dry_run,
        )
        result.files_uploaded.append(
            sbom_blob_path(
                connector_name=connector_name,
                version=version,
                store=store,
            ),
        )
        _log_progress("Uploaded SBOM to dedicated path: %s", sbom_gcs_uri)

    return result
